{#   Copyright (c) 2018 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
#}
# Configuration for Loadbalancer {{ loadbalancer.id }}
{% if loadbalancer.listener.enabled %}
# Configuration for Listener {{ udp_listener_id }}
{% else %}
# Listener {{ udp_listener_id }} is disabled
{% endif %}

{% block global_definitions %}{% endblock global_definitions %}

{% block proxies %}{% endblock proxies %}
